﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace TinyBravers {

	public class RotateMatrix {

		// After transpose we swap elements of column
		// one by one for finding left rotation of matrix
		// by 90 degree
		private static void reverseColumns(int[,] arr, int size) {
			int t1, t2;

			for (int i = 0; i < size; i++) {
				for (int j = 0, k = size - 1; j < k; j++, k--) {
					t1 = arr[j, i];
					t2 = arr[k, i];

					arr[j, i] = t2;
					arr[k, i] = t1;
					//swap(arr[j, i], arr[k, i]);
				}
			}
		}

		// Function for do transpose of matrix
		private static void transpose(int[,] arr, int size) {
			int t1, t2;

			for (int i = 0; i < size; i++) {
				for (int j = i; j < size; j++) {
					t1 = arr[i, j];
					t2 = arr[j, i];

					arr[i, j] = t2;
					arr[j, i] = t1;
					//swap(arr[i][j], arr[j][i]);
				}
			}
		}

		// Function for print matrix
		public static void printMatrix(int[,] arr, int size) {
			string cout = "";

			for (int i = 0; i < size; i++) {
				for (int j = 0; j < size; j++) {
					cout += arr[i, j] + " ";
				}
				cout += '\n';
			}

			TinyLogger.Log(cout);
		}

		// Function to anticlockwise rotate matrix
		// by 90 degree
		public static void rotate90(int[,] arr, int size) {
			transpose(arr, size);
			reverseColumns(arr, size);
		}

		// Squares

		// After transpose we swap elements of column
		// one by one for finding left rotation of matrix
		// by 90 degree
		private static void reverseColumns(TinySquare[,] arr, int size) {
			TinySquare t1, t2;

			for (int i = 0; i < size; i++) {
				for (int j = 0, k = size - 1; j < k; j++, k--) {
					t1 = arr[j, i];
					t2 = arr[k, i];

					arr[j, i] = t2;
					arr[k, i] = t1;
					//swap(arr[j, i], arr[k, i]);
				}
			}
		}

		// Function for do transpose of matrix
		private static void transpose(TinySquare[,] arr, int size) {
			TinySquare t1, t2;

			for (int i = 0; i < size; i++) {
				for (int j = i; j < size; j++) {
					t1 = arr[i, j];
					t2 = arr[j, i];

					arr[i, j] = t2;
					arr[j, i] = t1;
					//swap(arr[i][j], arr[j][i]);
				}
			}
		}

		// Function for print matrix
		public static void printMatrix(TinySquare[,] arr, int size) {
			string cout = "";

			for (int y = 0; y < size; y++) {
				for (int x = 0; x < size; x++) {
					cout += arr[x, y].type + " ";
				}
				cout += '\n';
			}

			TinyLogger.Log(cout);
		}

		// Function to rotate matrix
		// by 90 degree
		public static void rotate90(TinySquare[,] arr, int size) {
			printMatrix(arr, size);
			transpose(arr, size);
			reverseColumns(arr, size);
			printMatrix(arr, size);
		}
	}
}