﻿using System.Collections;
using System.Collections.Generic;
using UnityEngine;

namespace TinyBravers {

	/// <summary>
	/// This class does the A* pathfinding algorythm.
	/// </summary>
	public class TinyPathFinding {

		class TinyPath {
			public int G;
			public int H;
			public TinySquare current;
			public TinyPath parent;

			public TinyPath(int newG, int newH, TinySquare cSquare, TinyPath pSquare) {
				G = newG;
				H = newH;
				current = cSquare;
				parent = pSquare;
			}
		}

		List<TinyPath> openList = new List<TinyPath>(50);
		List<TinyPath> closedList = new List<TinyPath>(50);

		public TinySquare[] Pathfind(TinySquare initialPos, TinySquare destination) {
			TinySquare[] result;

			openList.Clear();
			closedList.Clear();

			//closedList.Add(new TinyPath(0, 0, initialPos, null));
			openList.Add(new TinyPath(0, 0, initialPos, null));

			while (openList.Count != 0) {
				//TinyLogger.Log("Tile: " + closedList[closedList.Count - 1].current.x + ", " + closedList[closedList.Count - 1].current.y);

				AddShortestToClosedList(destination);

				if (closedList[closedList.Count - 1].current == destination) {
					break;
				}

				ExploreToOpenList(closedList[closedList.Count - 1], destination);
			}

			// No path has been found
			if (closedList[closedList.Count - 1].current != destination) {
				return new TinySquare[0];
			}

			FillReturnArray(out result);

			return result;
		}

		void ExploreToOpenList(TinyPath current, TinySquare destination) {
			if (current.current.topSquare != null && current.current.topSquare.bWalkable && !ClosedListContains(current.current.topSquare)) {
				if (!OpenListContains(current.current.topSquare)) {
					openList.Add(CreatePathToSquare(current, current.current.topSquare, destination));
				} else {
					UpdatePathToSquare(current.current.topSquare, current);
				}
			}
			if (current.current.rightSquare != null && current.current.rightSquare.bWalkable && !ClosedListContains(current.current.rightSquare)) {
				if (!OpenListContains(current.current.rightSquare)) {
					openList.Add(CreatePathToSquare(current, current.current.rightSquare, destination));
				} else {
					UpdatePathToSquare(current.current.rightSquare, current);
				}
			}
			if (current.current.bottomSquare != null && current.current.bottomSquare.bWalkable && !ClosedListContains(current.current.bottomSquare)) {
				if (!OpenListContains(current.current.bottomSquare)) {
					openList.Add(CreatePathToSquare(current, current.current.bottomSquare, destination));
				} else {
					UpdatePathToSquare(current.current.bottomSquare, current);
				}
			}
			if (current.current.leftSquare != null && current.current.leftSquare.bWalkable && !ClosedListContains(current.current.leftSquare)) {
				if (!OpenListContains(current.current.leftSquare)) {
					openList.Add(CreatePathToSquare(current, current.current.leftSquare, destination));
				} else {
					UpdatePathToSquare(current.current.leftSquare, current);
				}
			}
		}

		void AddShortestToClosedList(TinySquare destination) {
			TinyPath path = null;
			int shortestF = int.MaxValue;

			for (int i = openList.Count - 1; i >= 0; i--) {
				if ((openList[i].G + openList[i].H < shortestF)) {
					shortestF = openList[i].G + openList[i].H;
					path = openList[i];
				}
				if (openList[i].current == destination) {
					path = openList[i];
					break;
				}
			}

			closedList.Add(path);
			openList.Remove(path);
		}

		void FillReturnArray(out TinySquare[] squares) {
			int size = 0;

			TinyPath current = closedList[closedList.Count - 1];

			while (current.parent != null) {
				current = current.parent;
				size++;
			}
			size++;

			squares = new TinySquare[size];

			current = closedList[closedList.Count - 1];

			while (current.parent != null) {
				size--;
				squares[size] = current.current;
				current = current.parent;
			}

			size--;
			squares[size] = current.current;
		}

		bool ClosedListContains(TinySquare square) {
			for (int i = 0; i < closedList.Count; i++) {
				if (closedList[i].current == square) {
					return true;
				}
			}

			return false;
		}

		bool OpenListContains(TinySquare square) {
			for (int i = 0; i < openList.Count; i++) {
				if (openList[i].current == square) {
					return true;
				}
			}

			return false;
		}

		TinyPath CreatePathToSquare(TinyPath parent, TinySquare current, TinySquare destination) {
			return new TinyPath(parent.G + 1, ManhattanDistance(current, destination), current, parent);
		}

		void UpdatePathToSquare(TinySquare squareToTest, TinyPath current) {
			TinyPath pathToUpdate = null;

			for (int i = 0; i < openList.Count; i++) {
				if (openList[i].current == squareToTest) {
					pathToUpdate = openList[i];
					break;
				}
			}

			if ((current.G + 1) < pathToUpdate.G) {
				pathToUpdate.G = (current.G + 1);
				pathToUpdate.parent = current;
			}
		}

		public static int ManhattanDistance(TinySquare initial, TinySquare destination) {
			return Mathf.Abs(destination.x - initial.x) + Mathf.Abs(destination.y - initial.y);
		}
	}
}